package opmanager

import (
	"context"
	"crypto/tls"
	"fmt"
	"io"
	"log"
	"net"
	"net/http"
	"net/url"
	"regexp"
	"strconv"
	"strings"
	"sync"
	"time"

	"github.com/it-novum/event-collectd/oitc"
	jsoniter "github.com/json-iterator/go"
)

var json = jsoniter.ConfigCompatibleWithStandardLibrary

type OpManagerHttpClient struct {
	ctx                   context.Context
	config                Config
	client                http.Client
	checkResultChannel    chan *oitc.CheckResult
	rawCommandChannel     chan *oitc.BrokerRawCommand
	cancelDowntimeChannel chan *oitc.CancelDowntime
	mapping               *oitc.OpenITCOCKPITMapping
	wg                    sync.WaitGroup
	shutdown              chan struct{}
}

// New creates a HTTP client to query the OpManager HTTP API v1
// https://www.manageengine.com/network-monitoring/help/rest-api-opmanager.html
// See api_examples.md for plain curl examples
func New(config Config) (*OpManagerHttpClient, error) {
	var proxyFunc func(*http.Request) (*url.URL, error) = nil
	if config.ProxyAddress != "" {
		proxyURL, err := url.Parse(config.ProxyAddress)
		if err != nil {
			return nil, err
		}
		proxyFunc = http.ProxyURL(proxyURL)
	}

	return &OpManagerHttpClient{
		config: config,
		client: http.Client{
			Timeout: 60 * time.Second,
			Transport: &http.Transport{
				TLSHandshakeTimeout: 10 * time.Second,
				TLSClientConfig: &tls.Config{
					InsecureSkipVerify: config.IgnoreSslCertificate == 1,
				},
				Proxy: proxyFunc,
				DialContext: (&net.Dialer{
					Timeout:   90 * time.Second,
					KeepAlive: 90 * time.Second,
				}).DialContext,
			},
		},
	}, nil
}

// Sends a GET request to the given API endpoint.
// T is a generic datatype and can be any Go struct such as Device or Interface or what ever the OpManager API returns to the given API endpoint
// This function loops through all pages and parses the json result
// Usage:
// var deviceResults PaginateResponse[Device] // Create a result strut with generic type Device
// allResults, err := get(c, "/listDevices", {}, 1, 1, &deviceResults) //Pass deviceResults as pointer (&) like you would to for json.Unmarshal()
// This function return allResults so it can loop through all pages and return all devices / interfaces at once
//
// Why is this not a method of OpManagerHttpClient?
// Because of Go:
// In Go, methods on types can't have type parameters. Type parameters are only allowed on functions, not methods.
// If you need a generic function and you want to associate it with a type (like a method), you can define the function
// outside of the type and pass the instance explicitly.
func get[T any](c *OpManagerHttpClient, endpoint string, queryParams map[string]string, limit int64, page int64, results *PaginateResponse[T]) (PaginateResponse[T], error) {
	var allResults PaginateResponse[T]

	baseUrl := c.config.ApiUrl + endpoint

	u, err := url.Parse(baseUrl)
	if err != nil {
		return allResults, err
	}

	loadData := true

	ctx, cancel := context.WithTimeout(c.ctx, 65*time.Second)
	defer cancel()

	for loadData {
		query := u.Query()
		query.Set("apiKey", c.config.ApiKey)
		query.Set("rows", strconv.FormatInt(limit, 10)) // Amount of records per result
		query.Set("page", strconv.FormatInt(page, 10))  // Page to load

		// Add dynamic parameters to query string
		for key, value := range queryParams {
			query.Set(key, value)
		}

		u.RawQuery = query.Encode()

		request, err := http.NewRequestWithContext(ctx, "GET", u.String(), nil)
		if err != nil {
			log.Println("[OpManager] NewRequestWithContext error: ", err)
			return allResults, err
		}

		// Send HTTP Request
		response, err := c.client.Do(request)
		if err != nil {
			log.Println("[OpManager] Server error: ", err)
			return allResults, err
		}
		defer response.Body.Close()

		if response.StatusCode != http.StatusOK {
			log.Println("[OpManager] Error - Non-OK status code:", response.Status)
			return allResults, err
		}

		// Read and parse the response body
		body, err := io.ReadAll(response.Body)
		if err != nil {
			log.Println("[OpManager] Error reading response body:", err)
			return allResults, err
		}

		// Check if the API returned an error (even if status code is 200 Ok)
		var resErr ResponseError
		if err := json.Unmarshal(body, &resErr); err != nil {
			// Unexpected response at all
			return allResults, err
		}

		// Check if the API returned an error
		if resErr.Error.Message != "" {
			return allResults, fmt.Errorf(resErr.Error.Message)
		}

		if err := json.Unmarshal(body, &results); err != nil {
			log.Println("[OpManager] Error wile parsing response json:", err)
			return allResults, err
		}

		// Copy all results into one singe variable
		// This is the same as a for loop (basically)
		allResults.Rows = append(allResults.Rows, results.Rows...)
		allResults.Page = results.Page
		allResults.Records = results.Records
		allResults.Total = results.Total

		if results.Total.ToInt64() == page {
			break
		}

		// Load next page
		page++
	}

	// Return all results
	return allResults, nil

}

func (c *OpManagerHttpClient) Close() {
	close(c.shutdown)
	c.wg.Wait()
}

func (c *OpManagerHttpClient) startApiPoller() error {
	defer c.wg.Done()

	// Golangs ticker does not fire immediately so we need a "fake" tick.
	ticker := time.NewTicker(time.Duration(c.config.PollingInterval) * time.Second)
	firstTick := make(chan struct{}, 1)
	firstTick <- struct{}{}

	defer ticker.Stop()
	defer close(firstTick)

	var err error
	for {

		select {

		case <-firstTick:
			// Fire the poller immediately
			log.Println("[OpManager] Initial poll of OpManager API")
			err = c.PollOpManagerApi()

			if err == context.Canceled {
				log.Println("[OpManager] Context canceled. Shutdown OpManager Poller")
				return nil
			}

			if err != nil {
				log.Println("[OpManager] OpManager Poller error:", err)
				log.Print("[OpManager] Trying to reconnect on next polling interval tick")
			}

		case <-ticker.C:
			// New Tick, poll the OpManager API
			if c.config.Verbose {
				log.Println("[OpManager] New Tick - poll the OpManager API")
			}
			err = c.PollOpManagerApi()

			if err == context.Canceled {
				log.Println("[OpManager] Context canceled. Shutdown OpManager Poller")
				return nil
			}

			if err != nil {
				log.Println("[OpManager] OpManager Poller error:", err)
				log.Print("[OpManager] Trying to reconnect on next polling interval tick")
			}

		case <-c.ctx.Done():
			// Close OpManager Poller
			return nil
		}

	}
}

func (c *OpManagerHttpClient) PollOpManagerApi() error {
	if c.config.Verbose {
		log.Println("[OpManager] Connecting to OpManager")
	}

	// Load all devices
	devices, err := c.GetAllDevices()
	if err != nil {
		log.Println("[OpManager] Error while loading devices")
		return err
	}

	// Load all interfaces
	interfacesGrouped, err := c.GetAllInterfaces()
	if err != nil {
		log.Println("[OpManager] Error while loading interfaces")
		return err
	}

	// Load all alerts
	alertsGrouped, err := c.GetAllAlerts()
	if err != nil {
		log.Println("[OpManager] Error while loading alerts")
		return err
	}

	monitorsGrouped, err := c.GetAllMonitors()
	if err != nil {
		log.Println("[OpManager] Error while loading monitors")
		return err
	}

	for _, device := range devices {
		// Find host UUID for given device

		hostUuid, err := c.mapping.GetHostUuidByHostIdentifier(device.DeviceName)
		if err != nil {
			if c.config.Verbose {
				log.Println("[OpManager] ", err)
			}
			continue
		}

		statusNum := device.StatusNum.ToInt64()
		deviceOutput := fmt.Sprintf(
			"Status: %v - Vendor: %v, Type: %v [%v]",
			device.StatusStr,
			device.VendorName,
			device.Type,
			device.Category,
		)

		var longHostOutput []string
		if statusNum != OPMANAGER_STATUS_CLEAR {
			// Check if we have any alerts for this host.
			if alerts, ok := alertsGrouped[device.DeviceName]; ok {
				for _, alert := range alerts {
					if alert.StatusNum != OPMANAGER_STATUS_CLEAR {
						// Filter empty alerts
						longHostOutput = append(longHostOutput, fmt.Sprintf(
							"[%v]: %v (%v)",
							alert.StatusStr,
							alert.Message,
							alert.PrettyTime,
						))
					}
				}
			}
		}

		// Send host check result to FIFO
		c.checkResultChannel <- &oitc.CheckResult{
			HostName:           hostUuid,
			ServiceDescription: "",
			Output:             deviceOutput,
			LongOutput:         strings.Join(longHostOutput[:], "\n"),
			PerfData:           "", // Not available in OpManager
			CheckType:          oitc.HOST_CHECK,
			ReturnCode:         c.mapOpManagerDeviceStateToOitcHostState(statusNum),
			StartTime:          time.Now().Unix(),
			EndTime:            time.Now().Unix(),
			EarlyTimeout:       0,
			Latency:            0.0,
			ExitedOk:           1,
		}

		// Loop through all interfaces of this device / host
		for _, iface := range interfacesGrouped[device.DeviceName] {
			serviceUuid, err := c.mapping.GetServiceUuidByHostIdentifierAndServiceIdentifier(device.DeviceName, iface.InterfaceName)
			if err != nil {
				if c.config.Verbose {
					log.Println("[OpManager] ", err)
				}
				continue
			}

			serviceOutput := fmt.Sprintf(
				"[%v] Rx Traffic: %v, Tx Traffic: %v, Admin Status: %v, Oper Status: %v, IP Address: %v, Type: %v ",
				iface.StatusStr,
				iface.InTraffic,
				iface.OutTraffic,
				iface.AdminStatus,
				iface.OperStatus,
				iface.IpAddress,
				iface.Iftype,
			)

			// Send service check result to FIFO (interface)
			c.checkResultChannel <- &oitc.CheckResult{
				HostName:           hostUuid,
				ServiceDescription: serviceUuid,
				Output:             serviceOutput,
				LongOutput:         "",
				PerfData:           c.getPerformanceDataAsString(iface),
				CheckType:          oitc.SERVICE_CHECK,
				ReturnCode:         c.mapOpManagerInterfaceStateToOitcServiceState(iface.StatusNum),
				StartTime:          time.Now().Unix(),
				EndTime:            time.Now().Unix(),
				EarlyTimeout:       0,
				Latency:            0.0,
				ExitedOk:           1,
			}
		}

		// Loop through all monitors of this device / host
		for _, monitor := range monitorsGrouped[device.DeviceName] {
			if monitor.PollingStatus != "0" || monitor.LastCollectedTime == "" || len(monitor.Data) == 0 {
				// "0" == Active
				continue
			}

			serviceUuid, err := c.mapping.GetServiceUuidByHostIdentifierAndServiceIdentifier(device.DeviceName, monitor.PollId)
			if err != nil {
				if c.config.Verbose {
					log.Println("[OpManager] ", err)
				}
				continue
			}

			currentMonitorState, err := strconv.ParseInt(monitor.MonSeverity, 10, 64)
			if err != nil {
				if c.config.Verbose {
					log.Println("[OpManager] ", err)
				}
				continue
			}

			serviceOutput := fmt.Sprintf("Last Polled value: %v %v. Last Polled at: %v", monitor.Data[0].Value.String(), monitor.Unit, monitor.LastCollectedTime)
			state := c.mapOpManagerInterfaceStateToOitcServiceState(currentMonitorState)
			perfdata := getPerformanceDataAsStringOfMonitor(&monitor)
			longOutput := make([]string, 0)

			if len(monitor.Data) > 1 {
				longOutput = append(longOutput, "[b]Instances:[/b]")
				for _, metric := range monitor.Data {
					if metric.Value.String() != "" {
						longOutput = append(longOutput, fmt.Sprintf("[b]%v:[/b] %v %v", metric.Instance, metric.Value.String(), monitor.Unit))
					}
				}
			}

			if currentMonitorState == OPMANAGER_STATUS_CLEAR {
				// This monitor is in a Ok state. OpManager has no real output for such monitors so we create our own one

				// Send service check result to FIFO (monitor)
				c.checkResultChannel <- &oitc.CheckResult{
					HostName:           hostUuid,
					ServiceDescription: serviceUuid,
					Output:             serviceOutput,
					LongOutput:         strings.Join(longOutput[:], "\n"),
					PerfData:           perfdata,
					CheckType:          oitc.SERVICE_CHECK,
					ReturnCode:         state,
					StartTime:          time.Now().Unix(),
					EndTime:            time.Now().Unix(),
					EarlyTimeout:       0,
					Latency:            0.0,
					ExitedOk:           1,
				}
			} else {
				// This monitor is in a Non-Ok State. Lets see if we can find an alert for this monitor
				// Lets try to find some alerts for this monitor.
				// Check if we have any alerts for this host.
				if alerts, ok := alertsGrouped[device.DeviceName]; ok {
					var alertForMonitor Alert
					foundAlert := false
					for _, alert := range alerts {

						// Get the ID out of the Alert Entity (77_RAMUSAGE)
						pattern := regexp.MustCompile(`^(\d+)_`)
						match := pattern.FindStringSubmatch(alert.Entity)
						if len(match) >= 2 {
							if match[1] == monitor.PollId {
								foundAlert = true
								alertForMonitor = alert
							}
						}
					}

					if foundAlert {
						serviceOutput = fmt.Sprintf("[%v]: %v (%v)", alertForMonitor.StatusStr, alertForMonitor.Message, alertForMonitor.PrettyTime)
						longOutput = append(longOutput, "[b]Alert:[/b]")
						longOutput = append(longOutput, serviceOutput)
					}

					// Send service check result to FIFO (monitor)
					c.checkResultChannel <- &oitc.CheckResult{
						HostName:           hostUuid,
						ServiceDescription: serviceUuid,
						Output:             serviceOutput,
						LongOutput:         strings.Join(longOutput[:], "\n"),
						PerfData:           perfdata,
						CheckType:          oitc.SERVICE_CHECK,
						ReturnCode:         state,
						StartTime:          time.Now().Unix(),
						EndTime:            time.Now().Unix(),
						EarlyTimeout:       0,
						Latency:            0.0,
						ExitedOk:           1,
					}
				}
			}

		}

	}

	return nil
}

func (c *OpManagerHttpClient) GetAllDevices() ([]Device, error) {
	var dst PaginateResponse[Device]

	params := map[string]string{"sortByColumn": "deviceName", "sortByType": "asc"}
	devices, err := get(c, "/api/json/v2/device/listDevices", params, 500, 1, &dst)
	if err != nil {
		return nil, err
	}
	return devices.Rows, nil
}

func (c *OpManagerHttpClient) GetAllInterfaces() (map[string][]Interface, error) {
	var dst PaginateResponse[Interface]

	params := map[string]string{"sortByColumn": "displayName", "sortByType": "asc", "isFluidic": "true"}
	interfaces, err := get(c, "/api/json/device/listInterfaces", params, 500, 1, &dst)
	if err != nil {
		return nil, err
	}

	// Group all interfaces by the host name
	var interfacesGrouped = make(map[string][]Interface)
	for _, iface := range interfaces.Rows {
		if _, ok := interfacesGrouped[iface.DeviceName]; !ok {
			// Create an empty array for the interfaces of this device / host
			interfacesGrouped[iface.DeviceName] = make([]Interface, 0)
		}
		interfacesGrouped[iface.DeviceName] = append(interfacesGrouped[iface.DeviceName], iface)
	}

	return interfacesGrouped, nil
}

func (c *OpManagerHttpClient) GetAllAlerts() (map[string][]Alert, error) {
	var dst PaginateResponse[Alert]

	// OpManager support recommends to add "&alertType=ActiveAlarms" - we should verify this
	// 12.11.2024 - [## 10166639 ##] Re: OpManager API call limit increase
	// 11.12.2024 - OpManager removed "sortByCategory": "true" from the API with version "12.8.270"
	params := map[string]string{"sortByColumn": "modTime", "sortByType": "asc", "isFluidic": "true", "isWebclient": "true"}

	alerts, err := get(c, "/api/json/alarm/listAlarms", params, 500, 1, &dst)
	if err != nil {
		return nil, err
	}

	// Group all alerts by the host name
	var alertsGrouped = make(map[string][]Alert)
	for _, alert := range alerts.Rows {
		if _, ok := alertsGrouped[alert.DeviceName]; !ok {
			// Create an empty array for the alerts of this device / host
			alertsGrouped[alert.DeviceName] = make([]Alert, 0)
		}
		alertsGrouped[alert.DeviceName] = append(alertsGrouped[alert.DeviceName], alert)
	}

	return alertsGrouped, nil
}

// This uses an undocumented API call we found via the "Monitors Health Summary" Dashboard Widget of OpManager
func (c *OpManagerHttpClient) GetAllMonitors() (map[string][]MonitorDetails, error) {
	var dst PaginateResponse[Device]

	params := map[string]string{"sortByColumn": "displayName", "sortByType": "asc", "getMonitors": "true"}
	devices, err := get(c, "/api/json/v2/device/listDevices", params, 500, 1, &dst)
	if err != nil {
		return nil, err
	}

	// ITC-3163
	// Load all Probes (only available on Enterprise Systems)
	// For now we ignore errors as Probes are optional
	probesmap := make(map[string]string)

	probes, _ := c.GetAllProbes()
	log.Println("[OpManager] Number of Probes: ", len(probes))
	for _, probe := range probes {
		probesmap[probe.Name] = probe.ProbeId
	}

	// Group all monitors by the host name
	var monitorsGroupedFiltered = make(map[string][]Monitor)
	var mapDevicesToProbeId = make(map[string]string)
	for _, device := range devices.Rows {
		// Filter all devices that do not have any monitors
		if len(device.Monitors) > 0 {
			if device.ProbeName != "" && len(probesmap) > 0 {
				// We are some sort of Enterprise OpManager
				if probeId, found := probesmap[device.ProbeName]; found {
					mapDevicesToProbeId[device.DeviceName] = probeId
				}
			}

			if _, ok := monitorsGroupedFiltered[device.DeviceName]; !ok {
				// Create an empty array for the interfaces of this device / host
				// len = 0 so we do not get empty values in our monitorsGrouped list but we set cap so we avoid memory allocations
				monitorsGroupedFiltered[device.DeviceName] = make([]Monitor, 0, len(device.Monitors))
			}

			//for _, monitor := range device.Monitors {
			//	monitorsGrouped[device.DeviceName] = append(monitorsGrouped[device.DeviceName], monitor)
			//}

			// This code does the same, as the for loop above but faster
			monitorsGroupedFiltered[device.DeviceName] = append(monitorsGroupedFiltered[device.DeviceName], device.Monitors...)
		}
	}

	var monitorsGrouped = make(map[string][]MonitorDetails)
	if len(monitorsGroupedFiltered) > 0 {
		// We have monitors at all
		// Now we need to run a GET request for each device to get the current value of the monitor :(
		for deviceName, monitors := range monitorsGroupedFiltered {

			// OpManager Enterprise requires the eeProbeID GET Parameter
			// Lets see if we have a ProbeId fro the current device
			probeId := ""
			if pId, found := mapDevicesToProbeId[deviceName]; found {
				probeId = pId
			}

			monitorDetails, err := c.getDeviceAssociatedMonitorsByDeviceName(deviceName, probeId)

			if err != nil {
				log.Println("[OpManager] Error while loading monitor details", err)
				continue
			}

			for _, monitorDetails := range monitorDetails.PerformanceMonitors.Monitors {
				//Only keep monitors with thresholds
				if monitorDetails.ThresholdEnabled != "true" {
					continue
				}

				if _, ok := monitorsGrouped[deviceName]; !ok {
					monitorsGrouped[deviceName] = make([]MonitorDetails, 0)
				}

				// Find this monitor in the first API result to copy over the MonSeverity value
				// Because the getDeviceAssociatedMonitors does not return the actual state of the monitor (╯°□°）╯︵ ┻━┻
				for _, monitor := range monitors {
					if monitor.PollID == monitorDetails.PollId {
						monitorDetails.MonSeverity = monitor.MonSeverity
					}
				}

				monitorsGrouped[deviceName] = append(monitorsGrouped[deviceName], monitorDetails)
			}
		}
	}

	return monitorsGrouped, nil
}

// ITC-3163
// OpManager Enterprise Systems can have multipe Probs. We need the eeProbeID to query GetAllMonitors on Enterprise Systems
// This API does not support pagination
func (c *OpManagerHttpClient) GetAllProbes() ([]Probe, error) {
	var allProbes []Probe

	baseUrl := c.config.ApiUrl + "/api/json/enterprise/listProbes"

	u, err := url.Parse(baseUrl)
	if err != nil {
		return allProbes, err
	}

	query := u.Query()
	query.Set("apiKey", c.config.ApiKey)

	u.RawQuery = query.Encode()

	request, err := http.NewRequestWithContext(c.ctx, "GET", u.String(), nil)
	if err != nil {
		return allProbes, err
	}

	// Send HTTP Request
	response, err := c.client.Do(request)
	if err != nil {
		log.Println("[OpManager] Server error: ", err)
		return allProbes, err
	}
	defer response.Body.Close()

	if response.StatusCode != http.StatusOK {
		log.Println("[OpManager] Error - Non-OK status code:", response.Status)
		return allProbes, err
	}

	// Read and parse the response body
	body, err := io.ReadAll(response.Body)
	if err != nil {
		log.Println("[OpManager] Error reading response body:", err)
		return allProbes, err
	}

	// Check if the API returned an error (even if status code is 200 Ok)
	var resErr ResponseError
	if err := json.Unmarshal(body, &resErr); err != nil {
		// Unexpacted response at all
		return allProbes, err
	}

	// Check if the API returned an error
	if resErr.Error.Message != "" {
		return allProbes, fmt.Errorf(resErr.Error.Message)
	}

	if err := json.Unmarshal(body, &allProbes); err != nil {
		log.Println("[OpManager] Error wile parsing response json:", err)
		return allProbes, err
	}

	// Return all results
	return allProbes, nil
}

func (c *OpManagerHttpClient) mapOpManagerDeviceStateToOitcHostState(statusNum int64) int64 {
	switch statusNum {

	case OPMANAGER_STATUS_CLEAR:
		return oitc.STATE_UP

	case OPMANAGER_STATUS_TROUBLE,
		OPMANAGER_STATUS_ATTENTION,
		OPMANAGER_STATUS_SERVICE_DOWN,
		OPMANAGER_STATUS_CRITICAL:
		return oitc.STATE_DOWN

	case OPMANAGER_STATUS_UNMANAGED:
		return oitc.STATE_UNREACHABLE
	}

	return oitc.STATE_UNREACHABLE
}

func (c *OpManagerHttpClient) mapOpManagerInterfaceStateToOitcServiceState(statusNum int64) int64 {
	switch statusNum {

	case OPMANAGER_STATUS_CLEAR:
		return oitc.STATE_OK

	case OPMANAGER_STATUS_TROUBLE,
		OPMANAGER_STATUS_ATTENTION:
		return oitc.STATE_WARNING

	case OPMANAGER_STATUS_SERVICE_DOWN,
		OPMANAGER_STATUS_CRITICAL:
		return oitc.STATE_CRITICAL

	case OPMANAGER_STATUS_UNMANAGED:
		return oitc.STATE_UNKNOWN
	}

	return oitc.STATE_UNKNOWN
}

func (c *OpManagerHttpClient) getPerformanceDataAsString(iface Interface) string {
	var perfdata []string

	rx, err := ParseInterfaceTraffic(iface.InTraffic)

	if err != nil || !rx.ParsedSuccessful {
		return ""
	}

	tx, err := ParseInterfaceTraffic(iface.OutTraffic)
	if err != nil || !rx.ParsedSuccessful {
		return ""
	}

	perfdata = append(perfdata, fmt.Sprintf("rx_traffic=%f%v;;;;", rx.OitcValue, rx.OitcUnit))
	perfdata = append(perfdata, fmt.Sprintf("rx_utilization=%f%%;;0;100", rx.Percentage))

	perfdata = append(perfdata, fmt.Sprintf("tx_traffic=%f%v;;;;", tx.OitcValue, tx.OitcUnit))
	perfdata = append(perfdata, fmt.Sprintf("tx_utilization=%f%%;;;0;100;", tx.Percentage))

	return strings.Join(perfdata[:], " ")
}

func (c *OpManagerHttpClient) getDeviceAssociatedMonitorsByDeviceName(deviceName string, probeId string) (DeviceAssociatedMonitors, error) {
	var allPerformanceMonitors DeviceAssociatedMonitors

	baseUrl := c.config.ApiUrl + "/api/json/device/getDeviceAssociatedMonitors"

	u, err := url.Parse(baseUrl)
	if err != nil {
		return allPerformanceMonitors, err
	}

	query := u.Query()
	query.Set("apiKey", c.config.ApiKey)
	query.Set("name", deviceName)

	// ITC-3163 OpManager Enterprise feature and undocumented
	if probeId != "" {
		query.Set("eeProbeID", probeId)
	}

	u.RawQuery = query.Encode()

	request, err := http.NewRequestWithContext(c.ctx, "GET", u.String(), nil)
	if err != nil {
		return allPerformanceMonitors, err
	}

	// Send HTTP Request
	response, err := c.client.Do(request)
	if err != nil {
		log.Println("[OpManager] Server error: ", err)
		return allPerformanceMonitors, err
	}
	defer response.Body.Close()

	if response.StatusCode != http.StatusOK {
		log.Println("[OpManager] Error - Non-OK status code:", response.Status)
		return allPerformanceMonitors, err
	}

	// Read and parse the response body
	body, err := io.ReadAll(response.Body)
	if err != nil {
		log.Println("[OpManager] Error reading response body:", err)
		return allPerformanceMonitors, err
	}

	if err := json.Unmarshal(body, &allPerformanceMonitors); err != nil {
		log.Println("[OpManager] Error wile parsing response json:", err)
		return allPerformanceMonitors, err
	}

	// Return all results
	return allPerformanceMonitors, nil

}

// Start reading from OpManager API returns immediately (SHOULD NOT RUN IN GOROUTINE)
// Due to there is no stream we can subscribe, we have to poll the API regularly
func (c *OpManagerHttpClient) Start(ctx context.Context, cr chan *oitc.CheckResult, raw chan *oitc.BrokerRawCommand, cancelDowntimeChannel chan *oitc.CancelDowntime, mapping *oitc.OpenITCOCKPITMapping) error {
	c.shutdown = make(chan struct{})
	c.checkResultChannel = cr
	c.rawCommandChannel = raw
	c.cancelDowntimeChannel = cancelDowntimeChannel
	c.mapping = mapping

	c.wg.Add(1)
	go func() {
		ctx, cancel := context.WithCancel(ctx)
		c.ctx = ctx

		defer func() {
			cancel()
			c.wg.Done()
		}()

		// Start a new thread (go func) that will poll the OpManager API regularly
		c.wg.Add(1)
		go c.startApiPoller()

		for {
			select {
			case <-ctx.Done():
				return

			case _, ok := <-c.shutdown:
				if !ok {
					// Channel c.shutdown got closed - close this OpManager poller
					log.Println("[OpManager] Shutdown OpManager poller")
					return
				}
			}

		}

	}()

	return nil
}
